from typing import Literal

from pydantic import BaseModel


TTSModels = Literal[
    "eleven_monolingual_v1",
    "eleven_multilingual_v1",
    "eleven_multilingual_v2",
    "eleven_turbo_v2",
    "eleven_turbo_v2_5",
]

TTSEncoding = Literal[
    "mp3_22050_32",
    "mp3_44100_32",
    "mp3_44100_64",
    "mp3_44100_96",
    "mp3_44100_128",
    "mp3_44100_192",
    "pcm_16000",
    "pcm_22050",
    "pcm_44100",
]

class VoiceSettings(BaseModel):
    stability: float  # [0.0 - 1.0]
    similarity_boost: float  # [0.0 - 1.0]
    style: float | None = None  # [0.0 - 1.0]
    use_speaker_boost: bool | None = False

class Voice(BaseModel):
    id: str
    name: str
    category: str
    settings: VoiceSettings | None = None


DEFAULT_VOICE = Voice(
    id="EXAVITQu4vr4xnSDxMaL",
    name="Bella",
    category="premade",
    settings=VoiceSettings(
        stability=0.71, similarity_boost=0.5, style=0.0, use_speaker_boost=True
    ),
)


from typing import Literal

DeepgramModels = Literal[
    "nova-general",
    "nova-phonecall",
    "nova-meeting",
    "nova-2-general",
    "nova-2-meeting",
    "nova-2-phonecall",
    "nova-2-finance",
    "nova-2-conversationalai",
    "nova-2-voicemail",
    "nova-2-video",
    "nova-2-medical",
    "nova-2-drivethru",
    "nova-2-automotive",
    "enhanced-general",
    "enhanced-meeting",
    "enhanced-phonecall",
    "enhanced-finance",
    "base",
    "meeting",
    "phonecall",
    "finance",
    "conversationalai",
    "voicemail",
    "video",
    "whisper-tiny",
    "whisper-base",
    "whisper-small",
    "whisper-medium",
    "whisper-large",
]

DeepgramLanguages = Literal[
    "zh",
    "zh-CN",
    "zh-TW",
    "da",
    "nl",
    "en",
    "en-US",
    "en-AU",
    "en-GB",
    "en-NZ",
    "en-IN",
    "fr",
    "fr-CA",
    "de",
    "hi",
    "hi-Latn",
    "pt",
    "pt-BR",
    "es",
    "es-419",
    "hi",
    "hi-Latn",
    "it",
    "ja",
    "ko",
    "no",
    "pl",
    "pt",
    "pt-BR",
    "es-LATAM",
    "sv",
    "ta",
    "taq",
    "uk",
    "tr",
    "sv",
    "id",
    "pt",
    "pt-BR",
    "ru",
    "th",
]



import os


def fetch_site_url():
    base_url = os.getenv("VERCEL_URL", "jay.so")
    if "http" not in base_url:
        if "localhost" in base_url:
            base_url = f"http://{base_url}"
        else:
            base_url = f"https://{base_url}"
    return base_url


def fetch_headers(api_key: str):
    api_protection_bypass_secret = os.getenv("VERCEL_AUTOMATION_BYPASS_SECRET")
    headers = {
        "Content-Type": "application/json",
        "X-API-Key": api_key,
    }
    if api_protection_bypass_secret:
        headers["x-vercel-protection-bypass"] = api_protection_bypass_secret
    return headers



from typing import Literal

TTSEncoding = Literal[
    "pcm_s16le",
    # Not yet supported
    # "pcm_f32le",
    # "pcm_mulaw",
    # "pcm_alaw",
]

TTSModels = Literal["sonic-english", "sonic-multilingual"]
TTSLanguages = Literal["en", "es", "fr", "de", "pt", "zh", "ja"]
TTSDefaultVoiceId = "c2ac25f9-ecc4-4f56-9095-651354df60c0"
TTSVoiceSpeed = Literal["fastest", "fast", "normal", "slow", "slowest"]
TTSVoiceEmotion = Literal[
    "anger:lowest",
    "anger:low",
    "anger",
    "anger:high",
    "anger:highest",
    "positivity:lowest",
    "positivity:low",
    "positivity",
    "positivity:high",
    "positivity:highest",
    "surprise:lowest",
    "surprise:low",
    "surprise",
    "surprise:high",
    "surprise:highest",
    "sadness:lowest",
    "sadness:low",
    "sadness",
    "sadness:high",
    "sadness:highest",
    "curiosity:lowest",
    "curiosity:low",
    "curiosity",
    "curiosity:high",
    "curiosity:highest",
]



from typing import List, Literal, Union


SpeechModels = Literal[
    "long",
    "short",
    "telephony",
    "medical_dictation",
    "medical_conversation",
    "chirp",
    "chirp_2",
]

SpeechLanguages = Literal[
    "en-US",
    "ja-JP",
    "en-IN",
    "en-GB",
    "hi-IN",
    "af-ZA",
    "sq-AL",
    "am-ET",
    "ar-EG",
    "hy-AM",
    "ast-ES",
    "az-AZ",
    "eu-ES",
    "be-BY",
    "bs-BA",
    "bg-BG",
    "my-MM",
    "ca-ES",
    "ceb-PH",
    "ckb-IQ",
    "zh-Hans-CN",
    "yue-Hant-HK",
    "zh-TW",
    "hr-HR",
    "cs-CZ",
    "da-DK",
    "nl-NL",
    "en-AU",
    "et-EE",
    "fil-PH",
    "fi-FI",
    "fr-CA",
    "fr-FR",
    "gl-ES",
    "ka-GE",
    "de-DE",
    "el-GR",
    "gu-IN",
    "ha-NG",
    "iw-IL",
    "hi-IN",
    "hu-HU",
    "is-IS",
    "id-ID",
    "it-IT",
    "ja-JP",
    "jv-ID",
    "kea-CV",
    "kam-KE",
    "kn-IN",
    "kk-KZ",
    "km-KH",
    "ko-KR",
    "ky-KG",
    "lo-LA",
    "lv-LV",
    "ln-CD",
    "lt-LT",
    "luo-KE",
    "lb-LU",
    "mk-MK",
    "no-NO",
    "pl-PL",
    "pt-BR",
    "pt-PT",
    "ro-RO",
    "ru-RU",
    "es-CO",
    "es-MX",
    "es-US",
    "th-TH",
    "tr-TR",
    "uk-UA",
    "vi-VN",
    "da-DK",
]

Gender = Literal["male", "female", "neutral"]

AudioEncoding = Literal["wav", "mp3", "ogg", "mulaw", "alaw", "linear16"]

LgType = Union[SpeechLanguages, str]
LanguageCode = Union[LgType, List[LgType]]


import os
import logging
import json
import requests
import inspect
import sentry_sdk
from pydantic import BaseModel
from logging import Logger
from typing import Awaitable, Callable, List, Any, Union
from dataclasses import field
from sentry_sdk.integrations.asyncio import AsyncioIntegration
from sentry_sdk.integrations.logging import LoggingIntegration
from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi import APIRouter, HTTPException, status
from fastapi.responses import StreamingResponse
from fastapi.encoders import jsonable_encoder
from jay_ai.token import generate_token
from jay_ai.cli_types import (
    LLMResponse,
    OnAgentStartedSpeakingInput,
    OnAgentStoppedSpeakingInput,
    OnUserStartedSpeakingInput,
    OnUserStoppedSpeakingInput,
    OnAgentMessageAddedInput,
    OnAgentInterruptedInput,
    OnFunctionCallsCollectedInput,
    OnFunctionCallsExecutedInput,
    OnUserMessageAddedInput,
    LLMResponseHandlerPayload,
    OnAgentMessageAddedPayload,
    OnAgentInterruptedPayload,
    OnAgentStartedSpeakingPayload,
    OnAgentStoppedSpeakingPayload,
    OnFunctionCallsCollectedPayload,
    OnFunctionCallsExecutedPayload,
    OnUserMessageAddedPayload,
    OnUserStartedSpeakingPayload,
    OnUserStoppedSpeakingPayload,
    OnAgentStartedSpeakingPayload,
    ConfigureSessionPayload,
    OnAgentMessageAddedInput,
    OnAgentInterruptedInput,
    OnFunctionCallsCollectedInput,
    OnFunctionCallsExecutedInput,
    OnUserMessageAddedInput,
    LLMResponseHandlerInput,
)
from jay_ai.cli_types import (
    SessionConfig,
    ConfigureSessionInput,
)
from jay_ai.utils import fetch_site_url, fetch_headers

logger = logging.getLogger(__name__)


async def _default_on_user_started_speaking(input: OnUserStartedSpeakingInput) -> None:
    pass


async def _default_on_user_stopped_speaking(input: OnUserStoppedSpeakingInput) -> None:
    pass


async def _default_on_agent_started_speaking(
    input: OnAgentStartedSpeakingInput,
) -> None:
    pass


async def _default_on_agent_stopped_speaking(
    input: OnAgentStoppedSpeakingInput,
) -> None:
    pass


async def _default_on_user_message_added(input: OnUserMessageAddedInput) -> None:
    pass


async def _default_on_agent_message_added(
    input: OnAgentMessageAddedInput,
) -> None:
    pass


async def _default_on_agent_interrupted(
    input: OnAgentInterruptedInput,
) -> None:
    pass


async def _default_on_function_calls_collected(
    input: OnFunctionCallsCollectedInput,
) -> None:
    pass


async def _default_on_function_calls_executed(
    input: OnFunctionCallsExecutedInput,
) -> None:
    pass


class Agent(BaseModel):
    id: str
    configure_session: Callable[[ConfigureSessionInput], Awaitable[SessionConfig]]
    llm_response_handler: Callable[[LLMResponseHandlerInput], Awaitable[LLMResponse]]
    tools: List[Callable[..., Awaitable[Any]]] = field(default_factory=list)
    on_user_started_speaking: Callable[
        [OnUserStartedSpeakingInput], Awaitable[None]
    ] = field(default=_default_on_user_started_speaking)
    on_user_stopped_speaking: Callable[
        [OnUserStoppedSpeakingInput], Awaitable[None]
    ] = field(default=_default_on_user_stopped_speaking)
    on_agent_started_speaking: Callable[
        [OnAgentStartedSpeakingInput], Awaitable[None]
    ] = field(default=_default_on_agent_started_speaking)
    on_agent_stopped_speaking: Callable[
        [OnAgentStoppedSpeakingInput], Awaitable[None]
    ] = field(default=_default_on_agent_stopped_speaking)
    on_user_message_added: Callable[[OnUserMessageAddedInput], Awaitable[None]] = field(
        default=_default_on_user_message_added
    )
    on_agent_message_added: Callable[[OnAgentMessageAddedInput], Awaitable[None]] = (
        field(default=_default_on_agent_message_added)
    )
    on_agent_interrupted: Callable[[OnAgentInterruptedInput], Awaitable[None]] = field(
        default=_default_on_agent_interrupted
    )
    on_function_calls_collected: Callable[
        [OnFunctionCallsCollectedInput], Awaitable[None]
    ] = field(default=_default_on_function_calls_collected)
    on_function_calls_executed: Callable[
        [OnFunctionCallsExecutedInput], Awaitable[None]
    ] = field(default=_default_on_function_calls_executed)


class ParsedAgent(Agent):
    def __post_init__(self) -> None:
        # Throw an error if any of the functions in `tools` is a lambda function. Named functions
        # are required because we use the function names to create API endpoints.
        for tool in self.tools:
            if inspect.isfunction(tool) and tool.__name__ == "<lambda>":
                raise ValueError(
                    "Lambda functions are not allowed in `tools`. Please define a named function instead."
                )

        # Find any duplicate function names
        tool_names = [tool.__name__ for tool in self.tools]
        duplicates = {name for name in tool_names if tool_names.count(name) > 1}
        if duplicates:
            raise ValueError(f"Duplicate tool names found: {', '.join(duplicates)}")

    def create_api(
        self, report_status: bool, skip_security_check: bool, deployment_id: str
    ):
        app = FastAPI()
        api_router = self.create_api_router()

        @app.middleware("http")
        async def middleware(request: Request, call_next):
            excluded_paths = ["/", "/health"]
            if request.url.path not in excluded_paths and skip_security_check == False:
                # Validate Agent API Key
                x_api_key = request.headers.get("X-API-Key")
                if x_api_key != os.getenv("JAY_INTERNAL__AGENT_API_KEY"):
                    raise HTTPException(status_code=401, detail="Invalid API key")
                # Report status
                if report_status:
                    report_service_status(deployment_id)
            response = await call_next(request)
            return response

        # Include our router
        app.include_router(api_router)

        # We only need this because we're calling /configureSession directly from the react app, but in production we should be calling
        # into the platform api which then forwards the request to the agent.
        # When we have that setup, we can remove this
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

        return app

    def create_production_api(self):
        deployment_id = os.getenv("JAY_INTERNAL__DEPLOYMENT_ID")
        sentry_dsn = os.getenv("SENTRY_DSN")
        sentry_env = os.getenv("SENTRY_ENVIRONMENT")
        if sentry_dsn and sentry_env:
            logger.info("Sentry integration initialized!")
            sentry_sdk.init(
                dsn=sentry_dsn,
                environment=sentry_env,
                # Sample rate for transactions (performance).
                traces_sample_rate=1.0,
                # Sample rate for exceptions / crashes.
                sample_rate=1.0,
                max_request_body_size="always",
                integrations=[
                    AsyncioIntegration(),
                    LoggingIntegration(level=logging.INFO, event_level=logging.ERROR),
                ],
            )
        else:
            logger.warning(
                "Sentry integration disabled due to missing Sentry configuration. Error logging with Sentry is highly recommended: https://docs.jay.so/error-monitoring"
            )

        return self.create_api(
            report_status=True, skip_security_check=False, deployment_id=deployment_id
        )

    def create_api_router(self):
        router = APIRouter()
        agent = self

        @router.get("/")
        def root():
            return {"message": "Service running"}

        @router.head("/")
        def root_head():
            return {"message": "Service running"}

        @router.get("/health")
        def health():
            return {"status": "healthy"}

        @router.post("/configureSession")
        async def configure_session_handler(payload: ConfigureSessionPayload):
            try:
                input_data = ConfigureSessionInput({"custom_data": payload.custom_data})
                session_config = await agent.configure_session(input_data)
                token = await generate_token(
                    session_config=session_config, agent_api_url=payload.agent_api_url
                )
                return {"token": token}
            except Exception as e:
                logger.exception("Error in /configureSession endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )

        @router.post("/agentResponse")
        async def llm_response_handler_endpoint(payload: LLMResponseHandlerPayload):
            try:
                input_data = LLMResponseHandlerInput(
                    {"session_data": payload.session_data, "messages": payload.messages}
                )

                response_stream = await agent.llm_response_handler(input_data)

                async def response_generator():
                    async for chunk in response_stream:
                        yield json.dumps(jsonable_encoder(chunk)).encode(
                            "utf-8"
                        ) + b"\n"

                return StreamingResponse(
                    response_generator(), media_type="application/json"
                )
            except Exception as e:
                logger.exception("Error in /agentResponse endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )

        @router.post("/userStartedSpeaking")
        async def user_started_speaking_endpoint(payload: OnUserStartedSpeakingPayload):
            input_data = OnUserStartedSpeakingInput(
                {"session_data": payload.session_data}
            )
            try:
                await agent.on_user_started_speaking(input_data)
            except Exception as e:
                logger.exception("Error in /userStartedSpeaking endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )
            return {"status": "ok"}

        @router.post("/userStoppedSpeaking")
        async def user_stopped_speaking_endpoint(payload: OnUserStoppedSpeakingPayload):
            input_data = OnUserStoppedSpeakingInput(
                {"session_data": payload.session_data}
            )
            try:
                await agent.on_user_stopped_speaking(input_data)
            except Exception as e:
                logger.exception("Error in /userStoppedSpeaking endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )
            return {"status": "ok"}

        @router.post("/agentStartedSpeaking")
        async def agent_started_speaking_endpoint(
            payload: OnAgentStartedSpeakingPayload,
        ):
            input_data = OnAgentStartedSpeakingInput(
                {"session_data": payload.session_data}
            )
            try:
                await agent.on_agent_started_speaking(input_data)
            except Exception as e:
                logger.exception("Error in /agentStartedSpeaking endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )
            return {"status": "ok"}

        @router.post("/agentStoppedSpeaking")
        async def agent_stopped_speaking_endpoint(
            payload: OnAgentStoppedSpeakingPayload,
        ):
            input_data = OnAgentStoppedSpeakingInput(
                {"session_data": payload.session_data}
            )
            try:
                await agent.on_agent_stopped_speaking(input_data)
            except Exception as e:
                logger.exception("Error in /agentStoppedSpeaking endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )
            return {"status": "ok"}

        @router.post("/userMessageAdded")
        async def user_message_added_endpoint(payload: OnUserMessageAddedPayload):
            try:
                input_data = OnUserMessageAddedInput(
                    {
                        "session_data": payload.session_data,
                        "message": payload.message,
                    }
                )

                await agent.on_user_message_added(input_data)
            except Exception as e:
                logger.exception("Error in /userMessageAdded endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )
            return {"status": "ok"}

        @router.post("/agentMessageAdded")
        async def agent_message_added_endpoint(
            payload: OnAgentMessageAddedPayload,
        ):
            try:
                input_data = OnAgentMessageAddedInput(
                    {
                        "session_data": payload.session_data,
                        "message": payload.message,
                    }
                )

                await agent.on_agent_message_added(input_data)
            except Exception as e:
                logger.exception("Error in /agentMessageAdded endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )
            return {"status": "ok"}

        @router.post("/agentInterrupted")
        async def agent_interrupted_endpoint(
            payload: OnAgentInterruptedPayload,
        ):
            try:
                input_data = OnAgentInterruptedInput(
                    {
                        "session_data": payload.session_data,
                        "message": payload.message,
                    }
                )

                await agent.on_agent_interrupted(input_data)
            except Exception as e:
                logger.exception("Error in /agentInterrupted endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )
            return {"status": "ok"}

        @router.post("/functionCallsCollected")
        async def function_calls_collected_endpoint(
            payload: OnFunctionCallsCollectedPayload,
        ):
            try:
                input_data = OnFunctionCallsCollectedInput(
                    {
                        "session_data": payload.session_data,
                        "function_calls": payload.function_calls,
                    }
                )

                await agent.on_function_calls_collected(input_data)
            except Exception as e:
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )
            return {"status": "ok"}

        @router.post("/functionCallsExecuted")
        async def function_calls_executed_endpoint(
            payload: OnFunctionCallsExecutedPayload,
        ):
            try:
                input_data = OnFunctionCallsExecutedInput(
                    {"session_data": payload.session_data, "results": payload.results}
                )

                await agent.on_function_calls_executed(input_data)
            except Exception as e:
                logger.exception("Error in /agentInterrupted endpoint")
                raise HTTPException(
                    status_code=status.HTTP_500_INTERNAL_SERVER_ERROR, detail=str(e)
                )
            return {"status": "ok"}

        def create_tool_endpoint(tool_func):
            async def _endpoint(payload: dict):
                try:
                    result = await tool_func(**payload)
                    return {"result": result}
                except Exception as e:
                    raise HTTPException(
                        status_code=status.HTTP_400_BAD_REQUEST,
                        detail=f"User logic error: {str(e)}",
                    )

            return _endpoint

        for tool_func in self.tools:
            func_name = tool_func.__name__
            endpoint_path = f"/tool/{func_name}"
            router.add_api_route(
                endpoint_path,
                create_tool_endpoint(tool_func),
                methods=["POST"],
                name=func_name,
            )

        return router


def fetch_report_status_payload(deployment_id: str):
    headers = fetch_headers(os.getenv("JAY_INTERNAL__AGENT_API_KEY"))
    url = f"{fetch_site_url()}/api/serviceStatus/report"
    payload = {
        "deployment_id": deployment_id,
    }

    return url, payload, headers


# Calls an API endpoint to report status of the service.
# This is used to implement smooth upgrade logic by decommissioning the agent api only after it has not received
# any requests after some time.
def report_service_status(deployment_id: str):
    url, payload, headers = fetch_report_status_payload(deployment_id)

    try:
        response = requests.post(url, json=payload, headers=headers)
        response.raise_for_status()
        logger.info(f"Response status: {response.status_code}")
        logger.info(f"Response data: {response.json()}")
    except requests.exceptions.HTTPError as http_err:
        logger.error(f"HTTP error occurred: {http_err}")
        logger.error(f"Response: {response.text}")
    except Exception as err:
        logger.exception(f"An unexpected error occurred: {err}")



from pathlib import Path
import shutil
import sys
import os
from typing import Callable, Optional, Union
import requests
from rich.console import Console
from rich.prompt import Prompt, Confirm
from rich.panel import Panel

from InquirerPy import inquirer
from InquirerPy.utils import InquirerPyStyle

from jay_ai.cli_types import LLMProvider, STTProvider, TTSProvider
from jay_ai.agent_template import get_agent_template
from jay_ai.utils import fetch_site_url, fetch_headers

console = Console()


def init_cli_command(getenv: Callable[[str, Optional[str]], Optional[str]] = os.getenv):
    def select_and_configure_provider(
        provider_enum: Union[type[STTProvider], type[LLMProvider], type[TTSProvider]],
        provider_display_name: str,
        env_var_inputs: dict,
        provider_api_keys: dict,
    ):
        display_labels = [provider.value.display_label for provider in provider_enum]

        chosen_label = inquirer.select(
            message=f"Select your {provider_display_name} provider",
            qmark="",
            instruction="(arrow keys to choose, then Enter)",
            pointer="❯",
            amark="",
            choices=display_labels,
            style=InquirerPyStyle(
                {
                    "question": "bold fg:#1E90FF",
                    "instruction": "italic",
                    "answer": "bold fg:#1E90FF",
                }
            ),
        ).execute()

        console.clear()

        chosen_provider = next(
            (p for p in provider_enum if p.value.display_label == chosen_label), None
        )
        for provider_env_var in chosen_provider.value.env_var_keys:
            key = provider_env_var.env_var_key
            if key not in env_var_inputs:
                existing_val = getenv(key, None)
                if existing_val:
                    console.print(f"Using the '{key}' that's already in .env")
                    env_var_inputs[key] = existing_val
                    provider_api_keys[key] = existing_val
                else:
                    # Ask the user for the new key
                    if provider_env_var.is_json:
                        console.print(
                            Panel.fit(
                                f"[bold]Enter {chosen_provider.value.display_label} JSON credentials. "
                                "Paste them below (single line):[/bold]"
                            ),
                            style="#1E90FF",
                        )
                    else:
                        console.print(
                            Panel.fit(
                                f"[bold]Enter {chosen_provider.value.display_label} API key:[/bold]"
                            ),
                            style="#1E90FF",
                        )
                    user_input = Prompt.ask(f"[bold]{key}[/bold]")
                    console.clear()
                    env_var_inputs[key] = user_input
                    provider_api_keys[key] = user_input

        return chosen_provider

    website_url = fetch_site_url()
    console.clear()

    # Check JAY_API_KEY
    jay_api_key = getenv("JAY_API_KEY", None)
    if jay_api_key:
        console.print("Using the JAY_API_KEY already defined in .env")
    else:
        console.print(
            Panel.fit(
                f"[bold]Retrieve your Jay API key at: [/bold][underline]{website_url}/dashboard/agent/settings[/underline]",
            ),
            style="#1E90FF",
        )
        jay_api_key = Prompt.ask("[bold]Enter your Jay API key[/bold]")
        console.clear()

    # Check JAY_DEVELOPER_ID
    jay_dev_id = getenv("JAY_DEVELOPER_ID", None)
    if jay_dev_id:
        console.print("Using the JAY_DEVELOPER_ID already defined in .env")
    else:
        console.print(
            Panel.fit(
                f"[bold]Retrieve your Jay Developer ID at: [/bold][underline]{website_url}/dashboard/agent/settings[/underline]",
            ),
            style="#1E90FF",
        )
        jay_dev_id = Prompt.ask("[bold]Enter your Jay Developer ID[/bold]")
        console.clear()

    # Check whether the user has Ngrok installed
    if shutil.which("ngrok") is None:
        console.print(
            "[red]Error: Ngrok is not installed. Please install it from https://ngrok.com/[/red]"
        )
        sys.exit(1)

    # We'll build a dictionary of every environment variable that the user enters.
    env_var_inputs = {"JAY_API_KEY": jay_api_key, "JAY_DEVELOPER_ID": jay_dev_id}

    # Let the user pick STT, LLM, TTS providers
    provider_api_keys = {}
    stt_provider = select_and_configure_provider(
        STTProvider, "speech-to-text", env_var_inputs, provider_api_keys
    )
    llm_provider = select_and_configure_provider(
        LLMProvider, "LLM", env_var_inputs, provider_api_keys
    )
    tts_provider = select_and_configure_provider(
        TTSProvider, "text-to-speech", env_var_inputs, provider_api_keys
    )

    console.print("[bold]Fetching agent ID from the dashboard...[/bold]")

    # Get the agent IDs from the database
    headers = fetch_headers(env_var_inputs["JAY_API_KEY"])
    try:
        resp = requests.get(f"{website_url}/api/agents/getAgentIds", headers=headers)
        if resp.status_code != 200:
            console.print(
                f"[red]Error retrieving agent IDs from {website_url}/api/agents/getAgentIds: "
                f"{resp.status_code} {resp.text}[/red]"
            )
            sys.exit(1)
        data = resp.json()
        agent_ids = data.get("agentIds", [])
        if len(agent_ids) == 0:
            console.print(
                "[red]Error: Could not find an agent ID. Please report this to the developers.[/red]"
            )
            sys.exit(1)
        elif len(agent_ids) > 1:
            console.print(
                "[red]Error: Exactly one agent ID is required, but we found more than one. You should update to the latest Jay package version. If the issue persists, please report it to the developers.[/red]"
            )
            sys.exit(1)
        agent_id = agent_ids[0]
    except Exception as e:
        console.print(f"[red]Error fetching agent IDs: {e}[/red]")
        sys.exit(1)

    # Create or overwrite agent
    agent_py_content = get_agent_template(
        agent_id=agent_id,
        stt_provider=stt_provider,
        llm_provider=llm_provider,
        tts_provider=tts_provider,
    )
    agent_file_path = Path("agent/main.py")
    if not agent_file_path.exists():
        agent_file_path.parent.mkdir(parents=True, exist_ok=True)
        with agent_file_path.open("w", encoding="utf-8") as f:
            f.write(agent_py_content)
        console.print(f"[green]Agent was created at: {str(agent_file_path)}.[/green]")
    else:
        existing = agent_file_path.read_text(encoding="utf-8")
        if existing.strip() != agent_py_content.strip():
            overwrite = Confirm.ask(
                f"[bold yellow]Agent already exists at: {str(agent_file_path)}. Overwrite?[/bold yellow]"
            )
            if overwrite:
                with agent_file_path.open("w", encoding="utf-8") as f:
                    f.write(agent_py_content)
                console.print("Agent file was overwritten.")
            else:
                console.print("Skipping overwrite of agent.")

    # Create or update the .env file
    dot_env_path = Path(".env")
    if dot_env_path.exists():
        lines_to_append = []

        for key, new_value in env_var_inputs.items():
            old_value = getenv(key, None)
            if old_value is None:
                # This key doesn't exist in .env (or system env), so add it
                lines_to_append.append(f"{key}={new_value}")

        if lines_to_append:
            to_append = ["", "# The following lines were added by `jay init`:"]
            to_append.extend(lines_to_append)
            with dot_env_path.open("a", encoding="utf-8") as f:
                f.write("\n".join(to_append) + "\n")
            console.print("[green]Appended new environment variables to .env[/green]")
    else:
        # Create a new .env file
        with dot_env_path.open("w", encoding="utf-8") as f:
            content_lines = [f"{k}={v}" for k, v in env_var_inputs.items()]
            f.write("# Environment variables for your Jay agent\n")
            f.write("\n".join(content_lines) + "\n")
        console.print("[green]Created a new .env file[/green]")

    # Fetch existing env var keys from DB
    try:
        resp = requests.get(f"{website_url}/api/envVars/get", headers=headers)
        if resp.status_code != 200:
            console.print(
                f"[red]Error retrieving environment variables from DB: "
                f"{resp.status_code} {resp.text}[/red]"
            )
            sys.exit(1)
        data = resp.json()
        existing_db_envvars = data.get("envVars", [])
        db_keys = [item["key"] for item in existing_db_envvars]
    except Exception as e:
        console.print(f"[red]Error calling envVars/get: {e}[/red]")
        sys.exit(1)

    # Determine which env vars are new or would be overwritten
    keys_to_save = [k for k in list(provider_api_keys.keys()) if k not in db_keys]
    keys_already_saved = [k for k in list(provider_api_keys.keys()) if k in db_keys]

    if not keys_to_save:
        console.print(
            "[yellow]No environment variables to store in Jay's database. Skipping DB update.[/yellow]"
        )
        console.print("[green]Initialization complete.[/green]")
        return

    if keys_already_saved:
        console.print(
            Panel.fit(
                f"[bold yellow]The following environment variable keys are already stored "
                f"in Jay's database and will not be overwritten:[/bold yellow]\n"
                f"[bold]{', '.join(keys_already_saved)}[/bold]\n\n"
                f"To overwrite them manually, go to:\n"
                f"[underline]{website_url}/dashboard/agent/settings[/underline]"
            )
        )

    console.print(
        Panel.fit(
            f"[bold]We will store the following environment variable keys in Jay's database:[/bold]\n"
            f"{', '.join(keys_to_save)}"
        )
    )
    proceed = Confirm.ask("[bold]Confirm?[/bold]")
    if proceed:
        # Save the new env vars
        vars_to_save = [{"key": k, "value": provider_api_keys[k]} for k in keys_to_save]

        payload = {"variables": vars_to_save}
        try:
            save_resp = requests.post(
                f"{website_url}/api/envVars/save", json=payload, headers=headers
            )
            if save_resp.status_code != 200:
                console.print(
                    f"[red]Error saving environment variables: "
                    f"{save_resp.status_code} - {save_resp.text}[/red]"
                )
                sys.exit(1)
        except Exception as e:
            console.print(f"[red]Error saving environment variables: {e}[/red]")
            sys.exit(1)

        console.print(
            Panel.fit(
                "[bold green]Your credentials have been stored in Jay's database. [/bold green]\n"
                f"You can view/modify them at [underline]{website_url}/dashboard/agent/settings[/underline].\n\n"
                "[bold green]Project successfully initialized![/bold green]"
            )
        )
    else:
        console.print(
            Panel.fit(
                f"[yellow]Understood. You can always update environment variables manually "
                f"at [underline]{website_url}/dashboard/agent/settings[/underline][/yellow]"
            )
        )
        console.print("[green]Initialization complete.[/green]")



import os
import importlib.util
import sys
import inspect
from jay_ai.agent import ParsedAgent


def load_parsed_agent(agent_path: str) -> ParsedAgent:
    agent_dir = os.path.dirname(agent_path)
    if agent_dir not in sys.path:
        sys.path.insert(0, agent_dir)

    try:
        functions_module = load_functions_module(agent_path)
    except Exception as e:
        print(f"Error loading functions module: {e}")
        sys.exit(1)

    return ParsedAgent(**functions_module.agent.model_dump())


def load_functions_module(module_path):
    spec = importlib.util.spec_from_file_location("", module_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



import json
import os
import httpx
from jay_ai.cli_types import (
    STT,
    TTS,
    VAD,
    STTProvider,
    SessionConfig,
    TTSProvider,
    VADProvider,
)
from jay_ai.utils import fetch_site_url, fetch_headers


async def generate_token(session_config: SessionConfig, agent_api_url: str) -> str:
    # Convert to dict so we can remove fields easily
    tts_kwargs_dict = session_config.tts.model_dump()
    tts_provider = None
    tts_credentials = None

    if isinstance(session_config.tts, TTS.ElevenLabs):
        tts_provider = TTSProvider.ELEVENLABS.value.label
        # ElevenLabs uses `api_key`
        tts_credentials = tts_kwargs_dict["api_key"]
        del tts_kwargs_dict["api_key"]
    elif isinstance(session_config.tts, TTS.OpenAI):
        tts_provider = TTSProvider.OPENAI.value.label
        # OpenAI uses `api_key`
        tts_credentials = tts_kwargs_dict["api_key"]
        del tts_kwargs_dict["api_key"]
    elif isinstance(session_config.tts, TTS.Google):
        tts_provider = TTSProvider.GOOGLE.value.label
        # Google uses `credentials` (dict) -> JSON string
        creds_dict = tts_kwargs_dict["credentials"]
        tts_credentials = json.dumps(creds_dict)
        del tts_kwargs_dict["credentials"]
    elif isinstance(session_config.tts, TTS.Azure):
        tts_provider = TTSProvider.AZURE.value.label
        # Azure uses `api_key`
        tts_credentials = tts_kwargs_dict["api_key"]
        del tts_kwargs_dict["api_key"]
    elif isinstance(session_config.tts, TTS.Deepgram):
        tts_provider = TTSProvider.DEEPGRAM.value.label
        # Deepgram TTS also uses `api_key`
        tts_credentials = tts_kwargs_dict["api_key"]
        del tts_kwargs_dict["api_key"]
    elif isinstance(session_config.tts, TTS.Cartesia):
        tts_provider = TTSProvider.CARTESIA.value.label
        # Cartesia uses `api_key`
        tts_credentials = tts_kwargs_dict["api_key"]
        del tts_kwargs_dict["api_key"]
    else:
        raise Exception(f"Unknown TTS provider: {session_config.tts}")

    # Convert back to JSON (for the payload)
    tts_kwargs = json.dumps(tts_kwargs_dict)

    # Do the same for STT
    stt_kwargs_dict = session_config.stt.model_dump()
    stt_provider = None
    stt_credentials = None

    if isinstance(session_config.stt, STT.OpenAI):
        stt_provider = STTProvider.OPENAI.value.label
        stt_credentials = stt_kwargs_dict["api_key"]
        del stt_kwargs_dict["api_key"]
    elif isinstance(session_config.stt, STT.Azure):
        stt_provider = STTProvider.AZURE.value.label
        stt_credentials = stt_kwargs_dict["api_key"]
        del stt_kwargs_dict["api_key"]
    elif isinstance(session_config.stt, STT.Deepgram):
        stt_provider = STTProvider.DEEPGRAM.value.label
        stt_credentials = stt_kwargs_dict["api_key"]
        del stt_kwargs_dict["api_key"]
    else:
        raise Exception(f"Unknown STT provider: {session_config.stt}")

    stt_kwargs = json.dumps(stt_kwargs_dict)

    # VAD
    vad_kwargs = session_config.vad.model_dump_json()
    vad_provider = None
    if isinstance(session_config.vad, VAD.Silero):
        vad_provider = VADProvider.SILERO.value.label
    else:
        raise Exception(f"Unknown VAD provider: {session_config.vad}")

    payload = {
        "messages": session_config.initial_messages,
        "tts_provider": tts_provider,
        "tts_kwargs": tts_kwargs,
        "tts_credentials": tts_credentials,
        "stt_provider": stt_provider,
        "stt_kwargs": stt_kwargs,
        "stt_credentials": stt_credentials,
        "vad_provider": vad_provider,
        "vad_kwargs": vad_kwargs,
        "session_data": session_config.session_data,
        "first_message": session_config.first_message,
        "allow_interruptions": session_config.allow_interruptions,
        "interrupt_time_threshold": session_config.interrupt_time_threshold,
        "interrupt_word_threshold": session_config.interrupt_word_threshold,
        "min_endpointing_delay": session_config.min_endpointing_delay,
        "max_nested_function_calls": session_config.max_nested_function_calls,
        "agent_api_url": agent_api_url,
    }
    endpoint = f"{fetch_site_url()}/api/generateToken"
    headers = fetch_headers(os.getenv("JAY_INTERNAL__AGENT_API_KEY"))

    async with httpx.AsyncClient() as client:
        try:
            response = await client.post(
                endpoint, json=payload, timeout=30.0, headers=headers
            )
            response.raise_for_status()
            response_json = response.json()
            token = response_json.get("token")
            if not token:
                raise ValueError(
                    "Token not found in the response from jay API. This should never happen."
                )

            return token

        except httpx.HTTPStatusError as http_err:
            raise Exception(
                f"HTTP error occurred: {http_err.response.status_code} - {http_err.response.text}"
            ) from http_err

        except httpx.RequestError as req_err:
            raise Exception(f"Request error occurred: {req_err}") from req_err

        except ValueError as val_err:
            raise Exception(f"Value error: {val_err}") from val_err



from jay_ai.cli_types import LLMProvider, STTProvider, TTSProvider


AGENT_CONFIG_TEMPLATE = """import os
from openai import AsyncOpenAI
from jay_ai import (
    VAD,
    STT,
    TTS,
    ConfigureSessionInput,
    SessionConfig,
    LLMResponseHandlerInput,
    Agent,
)

async def configure_session(input: ConfigureSessionInput):
    return SessionConfig(
        initial_messages=[],
        vad=VAD.Silero(),
        stt={stt_injection},
        tts={tts_injection},
    )


async def llm_response_handler(input: LLMResponseHandlerInput):
    client = {llm_injection}
    messages = input["messages"]
    completion = await client.chat.completions.create(
        model="gpt-4o",
        messages=messages,
        stream=True,
    )
    return completion


agent = Agent(
    id="{agent_id}",
    configure_session=configure_session,
    llm_response_handler=llm_response_handler
)
"""


def get_agent_template(
    agent_id: str,
    stt_provider: STTProvider,
    llm_provider: LLMProvider,
    tts_provider: TTSProvider,
) -> str:
    env_var_keys = tts_provider.value.env_var_keys + stt_provider.value.env_var_keys
    import_json = any(env_var.is_json for env_var in env_var_keys)

    file = AGENT_CONFIG_TEMPLATE.format(
        agent_id=agent_id,
        stt_injection=stt_provider.value.injection,
        llm_injection=llm_provider.value.injection,
        tts_injection=tts_provider.value.injection,
    )

    if import_json:
        file = "import json\n" + file

    return file



import sys
import os
import subprocess
import signal
import click
import httpx
import requests
import uvicorn
from dotenv import load_dotenv
import time
import socket
import webbrowser
import threading
import shutil
import importlib.resources
from pathlib import Path
from dotenv import load_dotenv
from fastapi import FastAPI, HTTPException, Request
from fastapi.staticfiles import StaticFiles
from jay_ai.loader import load_parsed_agent
from dotenv import load_dotenv
from fastapi.responses import JSONResponse
from pyngrok import ngrok
from jay_ai.utils import fetch_site_url, fetch_headers
from jay_ai.cli.init import init_cli_command

# ------------------------------------------------------------------------------
# Load Environment Variables
# ------------------------------------------------------------------------------
load_dotenv()

platform_url = fetch_site_url()
api_protection_bypass_secret = os.getenv("VERCEL_AUTOMATION_BYPASS_SECRET")
developer_id = os.getenv("JAY_DEVELOPER_ID")
api_key = os.getenv("JAY_API_KEY")

headers = fetch_headers(api_key)


def ensure_valid_agent_path(agent_path: str):
    if agent_path.endswith("main.py"):
        return
    else:
        raise Exception("Agent must be stored in a file named main.py")


def get_playground_path():
    """Determine the path to the playground directory."""
    dev_path = os.path.abspath("./services/playground/out")
    if os.path.exists(dev_path):
        return dev_path

    prod_path = os.path.join(os.path.dirname(__file__), "static/playground")
    if os.path.exists(prod_path):
        return prod_path

    click.secho(
        "Error: Playground directory not found in development or production locations.",
        fg="red",
        bold=True,
    )
    sys.exit(1)


async def call_get_token(request: Request) -> JSONResponse:
    request_body = await request.json()

    async with httpx.AsyncClient() as client:
        try:
            response = await client.post(
                f"{platform_url}/api/startSession",
                headers=headers,
                json=request_body,
            )

            response.raise_for_status()

        except httpx.HTTPStatusError as exc:
            raise HTTPException(
                status_code=exc.response.status_code,
                detail=f"Error response {exc.response.status_code}: {exc.response.text}",
            ) from exc

    return JSONResponse(status_code=response.status_code, content=response.json())


@click.group()
def cli():
    pass


@cli.command()
def init():
    """
    Initialize a new project.
    """

    init_cli_command()


@cli.command()
@click.option(
    "--agent",
    "agent_path",
    type=str,
    required=True,
    help="Path to the file containing your Agent.",
)
@click.option("--host", type=str, default="localhost", help="Host to run the agent on.")
@click.option("--port", type=int, default=8000, help="Port to run the agent on.")
@click.option(
    "--connect", "-c", is_flag=True, help="Automatically open the playground."
)
def run(agent_path, host, port, connect):
    """
    Run the playground with a local agent.
    """
    agent_path_abs = os.path.abspath(agent_path)
    if not os.path.isfile(agent_path_abs):
        click.secho(
            f"The functions path {agent_path_abs} is not a file.", fg="red", bold=True
        )
        sys.exit(1)

    if not api_key:
        raise Exception(
            "Did not detect a 'JAY_API_KEY'. Please define it in your .env file."
        )

    if not developer_id:
        raise Exception(
            "Did not detect a 'JAY_DEVELOPER_ID'. Please define it in your .env file."
        )

    os.environ["JAY_INTERNAL__AGENT_API_KEY"] = api_key

    ensure_valid_agent_path(agent_path_abs)

    print(f"Loading agent from: {agent_path_abs}")
    agent = load_parsed_agent(agent_path_abs)
    app = agent.create_api(
        report_status=False,
        skip_security_check=True,
        deployment_id="not-required-during-dev",
    )

    # Determine and serve the playground files
    playground_path = get_playground_path()
    app.mount(
        "/playground",
        StaticFiles(directory=playground_path, html=True),
        name="playground",
    )

    @app.post("/jay/getToken")
    async def get_token(request: Request):
        return await call_get_token(request)

    # Allows the playground to fetch the JAY_AGENT_ID specified by the user in their .env file
    @app.get("/jay/fetch_agent_id")
    def fetch_agent_id():
        return JSONResponse(
            content={
                "AGENT_ID": agent.id,
                "ENVIRONMENT": "development",
                "DEVELOPER_ID": developer_id,
            }
        )

    # Setup ngrok tunnel
    http_tunnel = ngrok.connect(addr=str(port))
    url = http_tunnel.public_url
    update_development_url = f"{platform_url}/api/updateDevelopmentUrl"
    payload = {
        "agent_id": agent.id,
        "developer_id": developer_id,
        "url": url,
    }
    response = requests.post(update_development_url, json=payload, headers=headers)
    if response.status_code != 200:
        print(response.text)
        response.raise_for_status()

    def cleanup(_a, _b):
        click.secho("Shutting down processes...", fg="yellow", bold=True)
        sys.exit(0)

    signal.signal(signal.SIGINT, cleanup)
    signal.signal(signal.SIGTERM, cleanup)

    click.secho(
        f"Starting Agent API server at http://{host}:{port}", fg="green", bold=True
    )
    playground_url = f"http://{host}:{port}/playground"
    click.secho(f"Playground is available at {playground_url}", fg="cyan")
    if connect:
        webbrowser.open(url=playground_url, new=2)

    try:
        uvicorn.run(app, host=host, port=port, log_level="debug")
    except KeyboardInterrupt:
        cleanup()


@cli.command()
@click.option(
    "--agent",
    "agent_path",
    type=str,
    required=True,
    help="Path to the file containing your Agent.",
)
@click.option(
    "--host", type=str, default="localhost", help="Host to run the playground on."
)
@click.option("--port", type=int, default=8000, help="Port to run the playground on.")
def connect(agent_path, host, port):
    """
    Connect to a deployed agent.
    """
    app = FastAPI()

    # Determine and serve the playground files
    playground_path = get_playground_path()
    app.mount(
        "/playground",
        StaticFiles(directory=playground_path, html=True),
        name="playground",
    )

    agent = load_parsed_agent(agent_path)

    @app.post("/jay/getToken")
    async def get_token(request: Request):
        return await call_get_token(request)

    # Allows the playground to fetch the JAY_AGENT_ID specified by the user
    @app.get("/jay/fetch_agent_id")
    def fetch_agent_id():
        return JSONResponse(
            content={
                "AGENT_ID": agent.id,
                "ENVIRONMENT": "production",
            }
        )

    def cleanup(_a, _b):
        click.secho("Shutting down processes...", fg="yellow", bold=True)
        sys.exit(0)

    signal.signal(signal.SIGINT, cleanup)
    signal.signal(signal.SIGTERM, cleanup)

    def start_uvicorn():
        uvicorn.run(app, host=host, port=port, log_level="debug")

    try:
        server_thread = threading.Thread(target=start_uvicorn, daemon=True)
        server_thread.start()

        url = f"http://{host}:{port}/playground"
        click.secho(f"Playground is available at {url}", fg="cyan")
        time.sleep(1)
        webbrowser.open(url=url, new=2)

        server_thread.join()
    except KeyboardInterrupt:
        cleanup()


@cli.command()
@click.option(
    "--agent",
    "agent_path",
    type=str,
    required=True,
    help="Path to the file containing your Agent.",
)
@click.option(
    "--requirement",
    "-r",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="Path to requirements.txt containing agent dependencies.",
)
def deploy(agent_path, requirement):
    """
    Build and deploy the agent.
    """

    # --------------------------------------------------------------------------
    # SETUP
    # --------------------------------------------------------------------------
    agent_path_abs = os.path.abspath(agent_path)
    if not os.path.isfile(agent_path_abs):
        click.secho(
            f"The functions path {agent_path_abs} is not a file.", fg="red", bold=True
        )
        sys.exit(1)

    ensure_valid_agent_path(agent_path_abs)

    # API endpoints
    create_deployment_endpoint = f"{platform_url}/api/deploy/create"
    trigger_build_endpoint = f"{platform_url}/api/deploy/build"
    status_url = f"{platform_url}/api/serviceStatus/fetch"
    cancel_deployment_endpoint = f"{platform_url}/api/deploy/cancel"

    agent = load_parsed_agent(agent_path_abs)

    # Attempt to get presigned URL and deployment ID
    try:
        create_deployment_payload = {"agent_id": agent.id}
        create_deployment_payload_resp = requests.post(
            create_deployment_endpoint, headers=headers, json=create_deployment_payload
        )
        if create_deployment_payload_resp.status_code != 200:
            click.secho(
                f"Failed to get presigned URL: {create_deployment_payload_resp.status_code} - {create_deployment_payload_resp.text}",
                fg="red",
                bold=True,
            )
            sys.exit(1)
        deployment_data = create_deployment_payload_resp.json()
        upload_url = deployment_data["presignedUrl"]
        deployment_id = deployment_data["deploymentId"]
    except requests.exceptions.RequestException as e:
        click.secho(f"Error triggering Build: {e}", fg="red", bold=True)
        sys.exit(1)

    def call_cancel_deployment():
        try:
            cancel_deployment_payload = {"deployment_id": deployment_id}
            cancel_deployment_resp = requests.post(
                cancel_deployment_endpoint,
                headers=headers,
                json=cancel_deployment_payload,
            )
            if cancel_deployment_resp.status_code == 200:
                print(f"Deployment canceled!")
                sys.exit(1)
        except requests.exceptions.RequestException as e:
            click.secho(f"Error canceling deployment: {e}", fg="red", bold=True)
            sys.exit(1)

    def signal_handler(signum, frame):
        click.secho(f"Canceling deployment: {deployment_id} and exiting gracefully...")
        call_cancel_deployment()
        sys.exit(0)

    # Register signal handlers
    signal.signal(signal.SIGINT, signal_handler)  # Handle Ctrl+C
    signal.signal(signal.SIGTERM, signal_handler)  # Handle termination

    click.secho(f"Generated deployment ID: {deployment_id}", fg="cyan", bold=True)
    agent_dir = os.path.dirname(agent_path)

    # Determine whether we use local jay package path
    local_package_path = os.getenv("JAY_INTERNAL__LOCAL_PACKAGE_PATH", None)

    click.secho(f"Building Docker image from source: {agent_dir}", fg="cyan", bold=True)

    # If local_package_path is set, copy the local jay files to .tmp_jay
    # ----------------------------------------------------------------------------
    # 0. Copy the local_package_path contents to a tmp directory if it is set.
    #    This allows us to install the local version of our package in the docker
    #    file during deployments to test new versions.
    # ----------------------------------------------------------------------------
    tmp_jay_dir = ".tmp_jay"
    target = "prod"
    if local_package_path:
        click.secho(
            f"Using local jay source: {local_package_path}", fg="yellow", bold=True
        )
        target = "dev"

        # Cleanup any existing .tmp_jay
        if os.path.exists(tmp_jay_dir):
            shutil.rmtree(tmp_jay_dir)

        os.makedirs(f"{tmp_jay_dir}/jay", exist_ok=True)

        # local_package_path might be relative or absolute
        jay_src = Path(local_package_path).resolve()
        if not jay_src.is_dir():
            click.secho(
                f"The local_package_path {jay_src} is not a valid directory.",
                fg="red",
                bold=True,
            )
            sys.exit(1)

        # Copy the jay code into .tmp_jay/jay
        shutil.copytree(str(jay_src), f"{tmp_jay_dir}/jay_ai", dirs_exist_ok=True)

        # Also copy pyproject.toml from parent of local_package_path
        pyproject_src = jay_src.parent / "pyproject.toml"
        if not pyproject_src.is_file():
            click.secho(
                f"Could not find pyproject.toml at {pyproject_src}", fg="red", bold=True
            )
            sys.exit(1)
        shutil.copy(str(pyproject_src), f"{tmp_jay_dir}/pyproject.toml")

    # Use importlib.resources to locate the Dockerfile
    with importlib.resources.path("jay_ai", "Dockerfile") as dockerfile_path:
        # --------------------------------------------------------------------------
        # 1. BUILD DOCKER IMAGE
        # --------------------------------------------------------------------------
        build_cmd = [
            "docker",
            "build",
            "--build-arg",
            f"AGENT_SRC_PATH={agent_dir}",
            "--build-arg",
            f"DEPLOYMENT_ID={deployment_id}",
            "--build-arg",
            f"REQUIREMENT={requirement}",
            "--target",
            target,
            "-t",
            "agent-api",
            "-f",
            str(dockerfile_path),
            "--platform",
            "linux/amd64",
            ".",
        ]
        # If dev, also pass the path .tmp_jay as LOCAL_PACKAGE_PATH
        if local_package_path:
            build_cmd.insert(6, f"--build-arg")
            build_cmd.insert(7, f"LOCAL_PACKAGE_PATH={tmp_jay_dir}")

        try:
            subprocess.run(build_cmd, check=True)
            click.secho("Docker image built successfully.", fg="green", bold=True)
        except subprocess.CalledProcessError as e:
            click.secho(f"Docker build failed: {e}", fg="red", bold=True)
            call_cancel_deployment()
            sys.exit(1)
        except Exception as e:
            click.secho(
                f"Docker build failed with unknown error: {e}", fg="red", bold=True
            )
            call_cancel_deployment()
            sys.exit(1)
        finally:
            # Cleanup the .tmp_jay dir if we created it
            if local_package_path and os.path.exists(tmp_jay_dir):
                shutil.rmtree(tmp_jay_dir, ignore_errors=True)

    # --------------------------------------------------------------------------
    # 2. RUN CONTAINER LOCALLY FOR HEALTH CHECK
    # --------------------------------------------------------------------------
    container_name = f"agent-api-test-{deployment_id}"
    host_port = get_free_port()
    try:
        run_cmd = [
            "docker",
            "run",
            "-d",
            "--name",
            container_name,
            "-p",
            f"{host_port}:10000",
            "--platform",
            "linux/amd64",
            "agent-api",
        ]
        click.secho(f"Testing Docker container...", fg="cyan")
        subprocess.run(run_cmd, check=True)
        click.secho(
            f"Docker container '{container_name}' started on port {host_port}.",
            fg="cyan",
        )

        # Health check
        wait_for_container_health(host_port, timeout=30)
        click.secho("Container is healthy (200 from /health).", fg="green", bold=True)

    except subprocess.CalledProcessError as e:
        click.secho(
            f"Error running Docker container health check: {e}", fg="red", bold=True
        )
        print_container_logs(container_name)
        call_cancel_deployment()
        sys.exit(1)
    except TimeoutError as e:
        click.secho(f"Health check failed: {e}", fg="red", bold=True)
        print_container_logs(container_name)
        call_cancel_deployment()
        sys.exit(1)
    except Exception as e:
        click.secho(
            f"Unknown error running Docker container health check: {e}",
            fg="red",
            bold=True,
        )
        print_container_logs(container_name)
        call_cancel_deployment()
        sys.exit(1)
    finally:
        # Cleanup the container
        cleanup_container(container_name)

    # --------------------------------------------------------------------------
    # 3. SAVE DOCKER IMAGE TO TARBALL
    # --------------------------------------------------------------------------
    temp_image_path = "/tmp/agent-api.tar"
    try:
        click.secho(f"Saving Docker image to {temp_image_path}", fg="cyan")
        subprocess.run(
            ["docker", "save", "-o", temp_image_path, "agent-api"], check=True
        )
    except Exception as e:
        click.secho(f"Error saving Docker image: {e}", fg="red", bold=True)
        call_cancel_deployment()
        sys.exit(1)

    # --------------------------------------------------------------------------
    # 4. GET PRESIGNED URL & UPLOAD TARBALL TO S3
    # --------------------------------------------------------------------------
    try:
        # Upload the tar file to S3
        with open(temp_image_path, "rb") as f:
            put_headers = {"Content-Type": "application/x-tar"}
            upload_resp = requests.put(upload_url, data=f, headers=put_headers)
        if upload_resp.status_code not in [200, 204]:
            click.secho(
                f"Failed to upload tarball to S3: {upload_resp.status_code} - {upload_resp.text}",
                fg="red",
                bold=True,
            )
            call_cancel_deployment()
            sys.exit(1)

        click.secho("Tarball uploaded to S3 successfully.", fg="green", bold=True)

    except Exception as e:
        click.secho(f"Error uploading Docker image tar to S3: {e}", fg="red", bold=True)
        call_cancel_deployment()
        sys.exit(1)
    finally:
        if os.path.exists(temp_image_path):
            os.remove(temp_image_path)
            click.secho(f"Removed temporary file {temp_image_path}.", fg="cyan")

    # --------------------------------------------------------------------------
    # 5. TRIGGER CODEBUILD
    # --------------------------------------------------------------------------
    try:
        trigger_payload = {"deployment_id": deployment_id}
        trigger_resp = requests.post(
            trigger_build_endpoint, json=trigger_payload, headers=headers
        )
        if trigger_resp.status_code != 200:
            click.secho(
                f"Failed to trigger Build: {trigger_resp.status_code} - {trigger_resp.text}",
                fg="red",
                bold=True,
            )
            call_cancel_deployment()
            sys.exit(1)
    except Exception as e:
        click.secho(f"Error triggering Build: {e}", fg="red", bold=True)
        call_cancel_deployment()
        sys.exit(1)

    # --------------------------------------------------------------------------
    # 6. POLL DEPLOYMENT STATUS
    # --------------------------------------------------------------------------
    payload = {"deployment_id": deployment_id}
    spinner = ["-", "\\", "|", "/"]
    idx = 0
    while True:
        try:
            response = requests.post(status_url, json=payload, headers=headers)
        except requests.exceptions.RequestException as e:
            click.secho(
                f"Error polling deployment status please report this to the developer: {e}",
                fg="red",
                bold=True,
            )
            sys.exit(1)

        if response.status_code == 200:
            status_data = response.json()
            status = status_data.get("deploymentStatus", "")
            if status == "active":
                click.secho(
                    "\nDeployment completed successfully.", fg="green", bold=True
                )
                break
            elif status == "failed":
                click.secho("Deployment failed.", fg="red", bold=True)
                sys.exit(1)

            # Show a spinning indicator
            click.echo(f"\rDeploying... {spinner[idx % len(spinner)]}", nl=False)
            idx += 1

        elif response.status_code == 404:
            click.secho(
                "Deployment not found. Check the deployment ID.", fg="red", bold=True
            )
            sys.exit(1)
        elif response.status_code == 400:
            click.secho(
                f"Invalid request checking deployment status. Status code: {response.status_code}, response: {response.json()}",
                fg="red",
                bold=True,
            )
            sys.exit(1)
        else:
            click.secho(
                f"Error fetching status: {response.status_code} - {response.text}",
                fg="red",
                bold=True,
            )
            sys.exit(1)

        time.sleep(2)

    click.secho("Agent deployment process completed.", fg="cyan", bold=True)


def get_free_port():
    """Find a free port on localhost."""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        return s.getsockname()[1]


def wait_for_container_health(port, timeout=30, interval=2):
    """Poll the /health endpoint until it's responsive or timeout is reached."""
    health_url = f"http://localhost:{port}/health"
    start_time = time.time()
    print(f"Starting health check at {health_url} (timeout={timeout}s)...")

    while True:
        try:
            response = requests.get(health_url)
            if response.status_code == 200:
                return
        except requests.exceptions.ConnectionError:
            pass  # Container not ready yet

        if time.time() - start_time > timeout:
            raise TimeoutError(
                f"Health check at {health_url} timed out after {timeout}s."
            )
        time.sleep(interval)


def cleanup_container(container_name):
    """Stop and remove a Docker container."""
    try:
        subprocess.run(
            ["docker", "stop", container_name],
            check=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        print(f"Stopped container '{container_name}'.")
        try:
            subprocess.run(
                ["docker", "rm", container_name],
                check=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
            print(f"Removed container '{container_name}'.")
        except subprocess.CalledProcessError:
            click.secho(
                f"Container '{container_name}' could not be removed or does not exist.",
                fg="red",
                bold=True,
            )
    except subprocess.CalledProcessError:
        click.secho(f"Container '{container_name}' was not running.", fg="yellow")


def print_container_logs(container_name: str):
    """
    Retrieve and print the container logs in red for debugging.
    """
    print(f"Retrieving logs for container '{container_name}'...")
    try:
        result = subprocess.run(
            ["docker", "logs", container_name],
            capture_output=True,
            text=True,
            check=False,
        )
        if result.stdout:
            click.secho("--- Container Logs (STDOUT) ---", fg="red", bold=True)
            click.secho(result.stdout, fg="red")
        if result.stderr:
            click.secho("--- Container Logs (STDERR) ---", fg="red", bold=True)
            click.secho(result.stderr, fg="red")
    except Exception as ex:
        click.secho(
            f"Failed to retrieve logs for container '{container_name}': {ex}",
            fg="red",
            bold=True,
        )


if __name__ == "__main__":
    cli()



from enum import Enum
from typing import Dict, List, Any, Literal, Optional, Tuple, Union
from typing_extensions import TypedDict
from pydantic import BaseModel
from typing import List
from jay_ai.plugins import azure, elevenlabs, openai, google, cartesia, deepgram


class ProviderEnvVarKey(BaseModel):
    env_var_key: str
    is_json: bool


class ProviderInfo(BaseModel):
    label: str
    display_label: str
    injection: str | None
    env_var_keys: list[ProviderEnvVarKey]


class TTSProvider(Enum):
    ELEVENLABS = ProviderInfo(
        display_label="ElevenLabs",
        label="elevenlabs",
        injection='TTS.ElevenLabs(api_key=os.environ["ELEVENLABS_API_KEY"])',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="ELEVENLABS_API_KEY", is_json=False)
        ],
    )
    GOOGLE = ProviderInfo(
        display_label="Google",
        label="google",
        injection='TTS.Google(credentials=json.loads(os.environ["GOOGLE_CREDENTIALS"]))',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="GOOGLE_CREDENTIALS", is_json=True)
        ],
    )
    OPENAI = ProviderInfo(
        display_label="OpenAI",
        label="openai",
        injection='TTS.OpenAI(api_key=os.environ["OPENAI_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="OPENAI_API_KEY", is_json=False)],
    )
    AZURE = ProviderInfo(
        display_label="Azure",
        label="azure",
        injection='TTS.Azure(api_key=os.environ["AZURE_API_KEY"], region=os.environ["AZURE_REGION"])',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="AZURE_API_KEY", is_json=False),
            ProviderEnvVarKey(env_var_key="AZURE_REGION", is_json=False),
        ],
    )
    CARTESIA = ProviderInfo(
        display_label="Cartesia",
        label="cartesia",
        injection='TTS.Cartesia(api_key=os.environ["CARTESIA_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="CARTESIA_API_KEY", is_json=False)],
    )
    DEEPGRAM = ProviderInfo(
        display_label="Deepgram",
        label="deepgram",
        injection='TTS.Deepgram(api_key=os.environ["DEEPGRAM_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="DEEPGRAM_API_KEY", is_json=False)],
    )


class STTProvider(Enum):
    DEEPGRAM = ProviderInfo(
        display_label="Deepgram",
        label="deepgram",
        injection='STT.Deepgram(api_key=os.environ["DEEPGRAM_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="DEEPGRAM_API_KEY", is_json=False)],
    )
    OPENAI = ProviderInfo(
        display_label="OpenAI",
        label="openai",
        injection='STT.OpenAI(api_key=os.environ["OPENAI_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="OPENAI_API_KEY", is_json=False)],
    )
    AZURE = ProviderInfo(
        display_label="Azure",
        label="azure",
        injection='STT.Azure(api_key=os.environ["AZURE_API_KEY"], region=os.environ["AZURE_REGION"])',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="AZURE_API_KEY", is_json=False),
            ProviderEnvVarKey(env_var_key="AZURE_REGION", is_json=False),
        ],
    )


class VADProvider(Enum):
    SILERO = ProviderInfo(
        display_label="Silero", label="silero", injection=None, env_var_keys=[]
    )


class LLMProvider(Enum):
    OPENAI = ProviderInfo(
        display_label="OpenAI",
        label="openai",
        injection='AsyncOpenAI(api_key=os.environ["OPENAI_API_KEY"])',
        env_var_keys=[ProviderEnvVarKey(env_var_key="OPENAI_API_KEY", is_json=False)],
    )
    OPENAI_COMPATIBLE = ProviderInfo(
        display_label="OpenAI Compatible",
        label="openai_compatible",
        injection='AsyncOpenAI(api_key=os.environ["OPENAI_COMPATIBLE_API_KEY"])',
        env_var_keys=[
            ProviderEnvVarKey(env_var_key="OPENAI_COMPATIBLE_API_KEY", is_json=False)
        ],
    )


class Message(TypedDict):
    content: str
    role: Literal["system", "user", "assistant", "tool"]
    name: str | None = None
    tool_call_id: str | None = None


class TTS:
    class ElevenLabs(BaseModel):
        api_key: str
        voice: elevenlabs.Voice = elevenlabs.DEFAULT_VOICE
        model: elevenlabs.TTSModels | str = "eleven_turbo_v2_5"
        encoding: elevenlabs.TTSEncoding = "mp3_22050_32"
        enable_ssml_parsing: bool = False
        chunk_length_schedule: list[int] = [80, 120, 200, 260]  # range is [50, 500]

    class OpenAI(BaseModel):
        api_key: str
        model: openai.TTSModels | str = "tts-1"
        voice: openai.TTSVoices | str = "alloy"
        speed: float = 1.0

    class Google(BaseModel):
        credentials: dict
        language: google.SpeechLanguages | str = "en-US"
        gender: google.Gender | str = "neutral"
        voice_name: str = ""  # Not required
        encoding: google.AudioEncoding | str = "linear16"
        sample_rate: int = 24000
        pitch: int = 0
        effects_profile_id: str = ""
        speaking_rate: float = 1.0

    class Azure(BaseModel):
        api_key: str
        region: str
        sample_rate: int = 24000
        voice: str | None = None
        language: str | None = None
        prosody: azure.ProsodyConfig | None = None
        endpoint_id: str | None = None

    class Deepgram(BaseModel):
        api_key: str
        model: str = "aura-asteria-en"
        encoding: str = "linear16"
        sample_rate: int = 24000

    class Cartesia(BaseModel):
        api_key: str
        model: cartesia.TTSModels | str = "sonic-english"
        language: cartesia.TTSLanguages | str = "en"
        encoding: cartesia.TTSEncoding = "pcm_s16le"
        voice: str | list[float] = cartesia.TTSDefaultVoiceId
        speed: cartesia.TTSVoiceSpeed | float | None = None
        emotion: list[cartesia.TTSVoiceEmotion | str] | None = None
        sample_rate: int = 24000


class STT:
    class Deepgram(BaseModel):
        api_key: str
        model: deepgram.DeepgramModels = "nova-2-general"
        language: deepgram.DeepgramLanguages = "en-US"
        interim_results: bool = True
        punctuate: bool = True
        smart_format: bool = True
        sample_rate: int = 16000
        no_delay: bool = True
        endpointing_ms: int = 25
        # enable filler words by default to improve turn detector accuracy
        filler_words: bool = True
        keywords: list[Tuple[str, float]] = []
        profanity_filter: bool = False

    class Azure(BaseModel):
        api_key: str
        region: str
        sample_rate: int = 16000
        num_channels: int = 1
        # Azure handles multiple languages and can auto-detect the language used. It requires the candidate set to be set.
        languages: list[str] = ["en-US"]

    class OpenAI(BaseModel):
        api_key: str
        language: str = "en"
        model: openai.WhisperModels | str = "whisper-1"


class VAD:
    class Silero(BaseModel):
        min_speech_duration: float = 0.05
        min_silence_duration: float = 0.55
        prefix_padding_duration: float = 0.5
        max_buffered_speech: float = 60.0
        activation_threshold: float = 0.5
        sample_rate: Literal[8000, 16000] = 16000


class SessionConfig(BaseModel):
    initial_messages: List[Message]
    vad: VAD.Silero
    stt: Union[STT.OpenAI, STT.Azure, STT.Deepgram]
    tts: Union[
        TTS.OpenAI,
        TTS.ElevenLabs,
        TTS.Google,
        TTS.Azure,
        TTS.Deepgram,
        TTS.Cartesia,
    ]
    session_data: dict[str, Any] = {}
    first_message: str | None = None
    allow_interruptions: bool = True
    interrupt_time_threshold: float = 0.5
    interrupt_word_threshold: int = 0
    min_endpointing_delay: float = 0.5
    max_nested_function_calls: int = 1


class Function(TypedDict):
    arguments: str
    name: Optional[str]


class FunctionCallInput(TypedDict):
    id: str
    function: Function
    type: Literal["function"]


class FunctionCallResult(TypedDict):
    content: str
    role: Literal["tool"]
    function: Function
    tool_call_id: str


class ConfigureSessionInput(TypedDict):
    custom_data: dict


class LLMResponseToolCall(TypedDict):
    index: int
    id: Optional[str]
    type: Optional[str]
    function: Optional[Function]


class LLMResponseDelta(TypedDict):
    role: Optional[str]
    content: Optional[str]
    tool_calls: Optional[List[LLMResponseToolCall]]


class LLMResponseHandlerInput(TypedDict):
    messages: List[Message]
    session_data: dict[str, Any]


class LLMResponseChoiceItem(TypedDict):
    index: int
    delta: LLMResponseDelta
    finish_reason: Optional[str]


class LLMResponseUsage(TypedDict):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class LLMResponse(TypedDict):
    id: str
    object: Literal["chat.completion.chunk"]
    created: int
    model: str
    choices: List[LLMResponseChoiceItem]
    usage: Optional[LLMResponseUsage]


class OnUserStartedSpeakingInput(TypedDict):
    session_data: dict[str, Any]


class OnUserStoppedSpeakingInput(TypedDict):
    session_data: dict[str, Any]


class OnAgentStartedSpeakingInput(TypedDict):
    session_data: dict[str, Any]


class OnAgentStoppedSpeakingInput(TypedDict):
    session_data: dict[str, Any]


class OnUserMessageAddedInput(TypedDict):
    message: Message
    session_data: dict[str, Any]


class OnAgentMessageAddedInput(TypedDict):
    message: Message
    session_data: dict[str, Any]


class OnAgentInterruptedInput(TypedDict):
    message: Message
    session_data: dict[str, Any]


class OnFunctionCallsCollectedInput(TypedDict):
    function_calls: List[FunctionCallInput]
    session_data: dict[str, Any]


class OnFunctionCallsExecutedInput(TypedDict):
    results: List[FunctionCallResult]
    session_data: dict[str, Any]


class LLMResponseHandlerPayload(BaseModel):
    session_data: dict[str, Any]
    messages: List[Dict[str, Any]]


class ConfigureSessionPayload(BaseModel):
    custom_data: dict[str, Any]
    agent_api_url: str


class OnUserStartedSpeakingPayload(BaseModel):
    session_data: dict[str, Any]


class OnUserStoppedSpeakingPayload(BaseModel):
    session_data: dict[str, Any]


class OnAgentStartedSpeakingPayload(BaseModel):
    session_data: dict[str, Any]


class OnAgentStoppedSpeakingPayload(BaseModel):
    session_data: dict[str, Any]


class OnUserMessageAddedPayload(BaseModel):
    session_data: dict[str, Any]
    message: Message


class OnAgentMessageAddedPayload(BaseModel):
    session_data: dict[str, Any]
    message: Message


class OnAgentInterruptedPayload(BaseModel):
    session_data: dict[str, Any]
    message: Message


class OnFunctionCallsCollectedPayload(BaseModel):
    session_data: dict[str, Any]
    function_calls: List[FunctionCallInput]


class OnFunctionCallsExecutedPayload(BaseModel):
    session_data: dict[str, Any]
    results: List[FunctionCallResult]


class ToolCallPayload(BaseModel):
    function_arguments: dict[str, Any]
